#include <httplib.h>
#include <iostream>
#include <thread>
#include <Generator.h>
#include <boost/process.hpp>
#include <filesystem>

//#include <wait.h>

#include "ModuleBase.h"
#include <httplib.h>
#include <memory>
#include <chrono>

using namespace httplib;
namespace mdd{

    using namespace boost::process;
    class ModuleHTTP : public ModuleBase {
    private:
        std::string _fname;
        std::string _id;
        int _port;
        std::unique_ptr<child> _child;

    protected:
        bool connect();
        std::string str_to_json(const std::string& input);
        state updateInputs();
        state updateOutputs();
        void updateLayout();

    public:
        ModuleHTTP();// std::string fname, std::string id, int port);
        ~ModuleHTTP();
        state update() override;
    };

    bool ModuleHTTP::connect(){
        Client cli(_id, _port);
        std::string body;
        auto res = cli.Get("/status",
                           [&](const char *data, size_t data_length) {
                               body.append(data, data_length);
                               return true;
                           });
        if(body.empty()){
            return false;
        }
        body =  std::string(R"()") + body;
        json status = json::parse(body.c_str());
        if (status["status"].get<std::string>() == "ready"){
            return true;
        }
        else{
            return false;
        }
    }

    std::string ModuleHTTP::str_to_json(const std::string& input){
        if(input.empty()){
            return input;
        }
        std::string str = input;
        size_t start_pos = 0;
        while((start_pos = str.find('[', start_pos)) != std::string::npos) {
            str.replace(start_pos, 1, "{");
            start_pos += 1;
        }
        start_pos = 0;
        while((start_pos = str.find(']', start_pos)) != std::string::npos) {
            str.replace(start_pos, 1, "}");
            start_pos += 1;
        }
        return str;
    }

    state ModuleHTTP::updateInputs() {
        Client cli(_id, _port);

        std::string body;
        auto res = cli.Get("/inputs",
                           [&](const char *data, size_t data_length) {
                               body.append(data, data_length);
                               return true;
                           });
        assert(res->body.empty());
        body =  std::string(R"()") + body;
        json inputs = json::parse(body.c_str());

        for(int j = 0; j < (int)getInputs().size()-(int)inputs.size(); ++j){
            pop_backInput();
        }
        for(size_t i=0; i < getInputs().size(); i++){
            getInput(i)->setType(inputs[i]["type"].get<std::string>());
            getInput(i)->setAppendix(i);
            getInput(i)->setValue() = inputs[i]["value"].get<std::vector<double>>();
        }
        for(int index = getInputs().size(); index < inputs.size(); ++index){
            if (inputs[index]["value"].is_array())
            {
                addInput(inputs[index]["type"].get<std::string>(), inputs[index]["value"].get<std::vector<double>>());
            }
            else {
                addInput(inputs[index]["type"].get<std::string>(), { inputs[index]["value"].get<double>() });
                std::cout << "Warning: Server expects single values, but mdd work with arrays!" << std::endl;
            }
        }
        return state::UNCHANGED;
    }

    state ModuleHTTP::updateOutputs() {
        Client cli(_id, _port);

        std::string body;
        auto res = cli.Get("/outputs",
                           [&](const char *data, size_t data_length) {
                               body.append(data, data_length);
                               return true;
                           });

        assert(res->body.empty());
        body =  std::string(R"()") + body;
        json outputs = json::parse(body);
        state output_state = state::UNCHANGED;
        for(int j = 0; j < (int)getOutputs().size()-(int)outputs.size(); ++j){
            pop_backOutput();
            output_state = state::CHANGED;
        }
        for(size_t i=0; i < getOutputs().size(); i++){
            //getOutput(i)->setType(outputs[i]["type"].get<std::string>());
            //getOutput(i)->setAppendix(i);
            getOutput(i)->setValue(outputs[i]["value"].get<std::vector<double>>());
        }
        for(int index = getOutputs().size(); index < outputs.size(); ++index){
            if (outputs[index]["value"].is_array()) {
                addOutput(outputs[index]["type"].get<std::string>(), outputs[index]["value"].get<std::vector<double>>());
            }
            else
            {
                addOutput(outputs[index]["type"].get<std::string>(), { outputs[index]["value"].get<double>() });
                std::cout << "Warning: Server delivers single values, but mdd work with arrays!" << std::endl;
            }
            output_state = state::CHANGED;
        }
        return output_state;
    }

    void ModuleHTTP::updateLayout(){
        updateInputs();
        updateOutputs();
    }

    ModuleHTTP::ModuleHTTP()// std::string fname, std::string id, int port):
      //  _fname(fname),
       // _id(id),
       // _port(port)
    {
       /* setType("HTTP");

        if(!fname.empty()){
            if (std::filesystem::exists(fname))
            {
                _child = std::make_unique<child>("python3 " + fname + " " + std::to_string(port));
            }
            else {
                std::cout << "ERROR Couldnt find: " << fname << std::endl;
            }
        }
        while(!connect()){
            std::this_thread::sleep_for(std::chrono::microseconds(500));
        }
        updateLayout();*/
    }

    ModuleHTTP::~ModuleHTTP()
    {
        if (_child != nullptr)
        {
            _child->terminate();
            _child->join();
        }
    }

    state ModuleHTTP::update() {
        Client cli(_id, _port);
        json new_inputs;
        for (int i = 0; i < getInputs().size(); ++i) {
            json input;
            input["value"] = getInput(i)->getValue();
            //std::cout << "ModuleHTTP::update: last: " << getInput(i)->getValue().back() << std::endl;
            new_inputs.push_back(input);
        }
        std::string content = new_inputs.dump();
        while (!connect()) {
            std::this_thread::sleep_for(std::chrono::microseconds(500));
        }
        //std::cout << "ModuleHTTP::update: " << content << std::endl;
        auto res = cli.Post("/update",content.size(),
                            [&](size_t offset, size_t length, DataSink &sink) {
                                sink.write(content.data() + offset, length);
                                return true; // return 'false' if you want to cancel the request.
                            },"application/json");
        while(!connect()){
            std::this_thread::sleep_for(std::chrono::microseconds(500));
        }
        return updateOutputs();
    }

}

ADD_GENERATOR(mdd::ModuleHTTP);